/*
 * Apache 2.0 Anti-Cracking Detection and Prevention Module v0.3
 * (C) 2004-2005 Joseph W. Benden, Sr., CCNA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * For the latest version: http://www.uglyboxindustries.com/
 *
 * Change Log:
 * Apr 24, 2005 - joe@thrallingpenguin.com
 * - Cleaned up code and released as GPL
 * Feb 22, 2004 - joe@thrallingpenguin.com
 * - Initial version
 *
 */

#include "apr_lib.h"

#define APR_WANT_STRFUNC
#define APR_WANT_BYTEFUNC
#include "apr_want.h"
#include "apr_strings.h"

#include "ap_config.h"

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_core.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#include "mysql.h"

typedef struct allowsS {
 char *ip;
} allows;

typedef struct myantihackstruct {
 int mysql;
 int mysql_connected;
 int level;
 int expires;
 int debug;
 char mysql_host[128];
 int mysql_port;
 MYSQL mydata;
 char mysql_db[128];
 char mysql_user[128];
 char mysql_pass[128];
 char mysql_table[128];
 apr_array_header_t *allowsArray;
} antihack_state;

/* Function Prototypes */
int mysql_do_close(antihack_state *state);
int mysql_do_connect(antihack_state *state);
unsigned long mysql_get_field0_ulong(antihack_state *state, char *sql);

/* Declare module */
module AP_MODULE_DECLARE_DATA mod_anticrack;

/* MYSQL Utility Functions */
int mysql_do_close(antihack_state *state)
{
 if(state->mysql_connected == 1)
  mysql_close(&state->mydata);
 state->mysql_connected = 0;
 return 0;
}

int mysql_do_connect(antihack_state *state)
{
 /* are we already connected? */
 if(state->mysql_connected == 1)
  return 0;

 if(mysql_init(&state->mydata) &&
    mysql_real_connect(&state->mydata, state->mysql_host, state->mysql_user,
      state->mysql_pass, NULL, state->mysql_port, NULL, 0) ) {
  if(mysql_select_db(&state->mydata, state->mysql_db) < 0) {
   /* unable to connect to database */
   ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL, "AntiCrack: Could not select database");
   mysql_close(&state->mydata);
   return -1;
  }
 } else {
  /* unable to connect to server */
  ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL, "AntiCrack: Could not connect to database server");
  mysql_close(&state->mydata);
  return -1;
 }

 /* we connected, so mark connect */
 state->mysql_connected = 1;
 return 0;
}

unsigned long mysql_get_field0_ulong(antihack_state *state, char *sql)
{
 MYSQL_RES	*res;
 MYSQL_ROW	row;
 unsigned long	ret = 0;

 if(mysql_do_connect(state)!=0)
  return -1;
 if(mysql_query(&state->mydata, sql)) {
  /* error while executing sql statement */
  ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL, "AntiCrack: SQL Error:");
  ret = -1;
 } else {
  if((res = mysql_use_result(&state->mydata)) == NULL) {
   /* no results were returned */
   ret = -1;
  } else {
   if((row=mysql_fetch_row(res)) != NULL) {
    if(row != NULL)
     ret = (unsigned long)atol(row[0]);
    else
     ret = 0;
    while((row=mysql_fetch_row(res))!=NULL)
     ;
   }
   mysql_free_result(res);
  }
 }
 return ret;
}

/* Module Functions */

static void *init_anticrack_module(apr_pool_t *p, char *dummy)
{
 /* Read our state data and connect to the database */
 antihack_state *state = (antihack_state *)apr_pcalloc(p, sizeof(antihack_state));

 state->mysql = 0;
 state->mysql_connected = 0;
 state->mysql_port = 3306;
 state->debug = 0;
 state->expires = (5 * 60); /* five minutes */
 strcpy(state->mysql_host,"localhost");
 strcpy(state->mysql_user,"root");
 strcpy(state->mysql_pass,"");
 strcpy(state->mysql_db,"anticrack");
 strcpy(state->mysql_table,"anticrack");
 state->allowsArray = apr_array_make(p, 1, sizeof(allows));

 return (void*)state;

}

static void *make_mod_anticrack_state(apr_pool_t *p, server_rec *s)
{
 /* Allocate state data */
 antihack_state *state = (antihack_state *)apr_pcalloc(p,
  sizeof(antihack_state));

 state->mysql = 0;

 return (void *)state;
}

static int anticrack_check(request_rec *orig)
{
 char sql[1024];
 unsigned long id = 0;
 allows *al;
 int i;
 antihack_state *state = (antihack_state*)ap_get_module_config(orig->per_dir_config,&mod_anticrack);

 /* module hasnt been enabled for this directory */
 /* Return decline, because elsewhere could deny access */
 if(state->mysql == 0)
	return DECLINED;

 /* Is the user in the always allow list? */
 for(i=0; i<state->allowsArray->nelts; ++i) {
  al = (allows *)state->allowsArray->elts;
   /* do we have a match? */
   if(strcmp(al[i].ip, orig->connection->remote_ip)==0)
    return OK;
 }

 /* do the database lookup */
 sprintf(sql,"SELECT id FROM %s WHERE ipaddr='%s' and counter>=%d and auth=0",
     state->mysql_table, orig->connection->remote_ip, state->level);
 if(state->debug == 1) {
  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, orig,
      "anticrack_logwatch: %s",sql);
 }
 if((id=mysql_get_field0_ulong(state,sql)) <= 0) {
  return OK;
 }
 return 403;
}

static int anticrack_logwatch(request_rec *r)
{
 char sql[1024];
 unsigned long id;
 antihack_state *state = (antihack_state*)ap_get_module_config(r->per_dir_config,&mod_anticrack);

 if(state->debug == 1) {
  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
     "anticrack_logwatch: HTTP_CODE=%d",
     r->status);
 }
 if( state->mysql == 0 ) {
  /* this means that we are not enabled in this directory or for this */
  /* request. */
  return OK;
 }
 /* If status == 200 they authenticated already so dont flag them as hacks */
 if( r->status == 200 ) {
  sprintf(sql,"SELECT id FROM %s WHERE ipaddr='%s'",
   state->mysql_table, r->connection->remote_ip);
  if((id=mysql_get_field0_ulong(state,sql))<1) {
   sprintf(sql, "INSERT INTO %s SET lastattempt=NOW(), auth=1, counter=1, ipaddr='%s'",
     state->mysql_table, r->connection->remote_ip);
  } else {
   sprintf(sql, "UPDATE %s SET lastattempt=NOW(), counter=counter+1, auth=1 WHERE id=%ld",
     state->mysql_table, id);
  }
  mysql_query(&state->mydata,sql);  
 }
 if( r->status == 401 || r->status == 403 ) {
  /* Remove older entries from the table */
  sprintf(sql,"DELETE FROM %s WHERE counter<%d AND (noRemove=0 OR auth=1) AND UNIX_TIMESTAMP(lastattempt)<=UNIX_TIMESTAMP(NOW() - %d)",
    state->mysql_table, state->level, state->expires);
  mysql_query(&state->mydata,sql);
  if(state->debug == 1) {
   ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
     "anticrack_logwatch: %s",sql);
  }

  /* We've encountered a code that deserves to be watched. */
  sprintf(sql, "SELECT id FROM %s WHERE ipaddr='%s'",
   state->mysql_table, r->connection->remote_ip);
  if(state->debug == 1) {
   ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
     "anticrack_logwatch: %s",sql);
  }
  if((id=mysql_get_field0_ulong(state,sql))<1) {
   /* insert new record */
   sprintf(sql, "INSERT INTO %s SET lastattempt=NOW(), counter=1, ipaddr='%s'",
     state->mysql_table, r->connection->remote_ip);
  } else {
   /* update existing record */
   sprintf(sql,"UPDATE %s SET lastattempt=NOW(), counter=counter+1 WHERE id=%ld",
     state->mysql_table, id);
  }
  mysql_query(&state->mydata,sql);
  if(state->debug == 1) {
   ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
     "anticrack_logwatch: %s",sql);
  }
  /* solidify the existing result set */
  sprintf(sql,"UPDATE %s SET noRemove=1, lastattempt=lastattempt WHERE counter>=%d and noRemove=0",
    state->mysql_table, state->level);
  mysql_query(&state->mydata,sql);
  if(state->debug == 1) {
   ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
     "anticrack_logwatch: %s",sql);
  }
 }

 return OK;
}

static const char *antihack_set_host(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 /* arg - contains the hostname */
 state->mysql = 1;
 strncpy(state->mysql_host, arg, 128);

 return NULL;
};

static const char *antihack_set_db(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 strncpy(state->mysql_db, arg, 128);

 return NULL;
};

static const char *antihack_set_user(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 strncpy(state->mysql_user, arg, 128);

 return NULL;
};

static const char *antihack_set_pass(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 strncpy(state->mysql_pass, arg, 128);

 return NULL;
};

static const char *antihack_set_table(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 strncpy(state->mysql_table, arg, 128);

 return NULL;
};

static const char *antihack_set_level(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 state->level = atol(arg);

 return NULL;
};

static const char *antihack_set_expires(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 state->expires = atol(arg);

 return NULL;
};

static const char *antihack_set_debug(cmd_parms *cmd, void *dv, const char *arg)
{
 antihack_state *state = (antihack_state*)dv;
 if(strcmp(arg,"on")==0 || atol(arg)>0)
  state->debug = 1;

 return NULL;
};

static const char *antihack_add_allow(cmd_parms *cmd, void *dv, const char *arg)
{
 allows *pa = NULL;
 antihack_state *state = (antihack_state*)dv;
 char *where = apr_pstrdup(cmd->pool, arg);

 pa = apr_array_push(state->allowsArray);
 pa->ip = where;
 
 return NULL;
};

static const command_rec access_cmds[] = 
{
 AP_INIT_TAKE1("AntiCrackHost", antihack_set_host, NULL, OR_LIMIT,
  "set the mysql hostname for anticrack"),
 AP_INIT_TAKE1("AntiCrackDB", antihack_set_db, NULL, OR_LIMIT,
  "set the mysql database name used for anticrack"),
 AP_INIT_TAKE1("AntiCrackUser", antihack_set_user, NULL, OR_LIMIT,
  "set the mysql username used for anticrack"),
 AP_INIT_TAKE1("AntiCrackPass", antihack_set_pass, NULL, OR_LIMIT,
  "set the mysql password used for anticrack"),
 AP_INIT_TAKE1("AntiCrackTable", antihack_set_table, NULL, OR_LIMIT,
  "set the mysql table used for anticrack"),
 AP_INIT_TAKE1("AntiCrackLevel", antihack_set_level, NULL, OR_LIMIT,
  "set the level at which access denies become permanate"),
 AP_INIT_TAKE1("AntiCrackExpires", antihack_set_expires, NULL, OR_LIMIT,
  "sets the number of seconds for a crack attempt to remain in the table"),
 AP_INIT_TAKE1("AntiCrackDebug", antihack_set_debug, NULL, OR_LIMIT,
  "enables the debugging mode of operation"),
 AP_INIT_TAKE1("AntiCrackPermit", antihack_add_allow, NULL, OR_LIMIT,
  "allows an IP address access always"),
 {NULL}
};

static void register_hooks(apr_pool_t *p)
{
 ap_hook_access_checker(anticrack_check,NULL,NULL,APR_HOOK_MIDDLE);
 ap_hook_log_transaction(anticrack_logwatch,NULL,NULL,APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA mod_anticrack=
{
 STANDARD20_MODULE_STUFF,
 init_anticrack_module,
 NULL,				/* create per-dir config */
 NULL,				/* merge per-dir config */
 /*make_mod_anticrack_state,*/
 NULL,				/* merge server config */
 access_cmds,				/* command table */
 register_hooks
};

